'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;
  var template = _ref.template;

  var declarationTemplate = template('\n    var NAME = VALUE;\n  ');
  return {
    visitor: {
      Function: function Function(path) {
        if (t.isProgram(path.scope.parent.block)) {
          // Nowhere we can move this function
          return;
        }
        if (path.getData('unhoistable')) return;
        // Determine the outermost scope we can move to,
        // which is the innermost scope we are actually referencing (if any),
        // or the global scope.
        var innerScope = new InnerScopeVisitor();
        path.traverse(innerScope);
        if (innerScope.seenThisExpression && t.isArrowFunctionExpression(path.node)) return;
        var referencedScopes = uniqueScopes(innerScope.referencedScopes || []);
        var targetScope = deepestScopeOf(path, referencedScopes.concat(path.scope.getProgramParent()).filter(function (scope) {
          return scope !== path.scope;
        }));
        if (!targetScope || targetScope === path.scope.parent) {
          return;
        }
        if (path.node.id) {
          var binding = path.scope.getBinding(path.node.id.name);
          var isUnsafeUse = function isUnsafeUse(usePath) {
            if (usePath.parent.type !== 'MemberExpression') return false;
            if (usePath.key !== 'object') return false;
            while (usePath = usePath.parentPath) {
              // eslint-disable-line no-cond-assign
              if (usePath.parent.type === 'UpdateExpression' && usePath.key === 'argument') {
                return true;
              }
              if (usePath.parent.type === 'AssignmentExpression' && usePath.key === 'left') {
                return true;
              }
              if (!t.isLVal(usePath.node)) {
                break;
              }
            }
            return false;
          };
          var uses = binding.referencePaths.filter(function (refPath) {
            return refPath !== path && t.isIdentifier(refPath);
          });

          if (uses.length) {
            if (uses.some(isUnsafeUse)) {
              return;
            }
          }
        }

        // FunctionDeclaration needs some extra processing to be compatible with Babel's hoister:
        // 1) Replace the node with a VariableDeclaration initialized with a FunctionExpression.
        // 2) Invoke hoist() on the expression.
        // 3) Hoist the declaration itself within its block so the function is correctly available
        //    even before its original point of declaration.

        var declaratorPath = null;
        var declaratorBackupNode = null;
        if (t.isFunctionDeclaration(path.node)) {
          declaratorPath = path;
          declaratorBackupNode = declaratorPath.node;
          path.replaceWith(declarationTemplate({
            NAME: path.node.id,
            VALUE: Object.assign({}, path.node, { type: 'FunctionExpression' })
          }));
          path = declaratorPath.get('declarations.0.init');
        }

        // Babel's hoister will give the outer declaration a meaningless id like "_ref" :(
        // We generate a meaningful one instead, and we do it *before* hoisting so as to completely
        // ignore the hoister's id (in case ours happen to collide with it).
        var id = targetScope.generateUidIdentifierBasedOnNode(path.node.id || path.node, 'hoistedAnonymousFunc');
        path.hoist(targetScope);
        // fix up the temporary hoisted name to our meaningful id
        var tempName = path.node.name;
        if (!tempName) {
          // hoisting did not succeed
          if (declaratorBackupNode && declaratorPath) {
            declaratorPath.setData('unhoistable', true);
            declaratorPath.replaceWith(declaratorBackupNode);
          }
          return;
        }
        targetScope.crawl(); // so the next line can find the newly inserted binding
        var tempBinding = path.scope.getBinding(tempName);
        tempBinding.path.get('id').replaceWith(id);
        path.replaceWith(id);
        path.scope.crawl(); // so we see the renamed reference in further processing
        tempBinding.path.scope.crawl(); // so the scope holding the temp binding is fully updated too

        // fix up the inner VariableDeclarator's position
        if (declaratorPath) {
          var _declaratorPath = declaratorPath;
          var node = _declaratorPath.node;

          declaratorPath.remove();
          declaratorPath.scope.getFunctionParent().path.get('body').unshiftContainer('body', node);
        }
      }
    }
  };
};

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var InnerScopeVisitor = function InnerScopeVisitor() {
  var _this = this;

  _classCallCheck(this, InnerScopeVisitor);

  this.ReferencedIdentifier = function (path) {
    var binding = path.scope.getBinding(path.node.name);
    if (binding) {
      if (!_this.referencedScopes) {
        _this.referencedScopes = [];
      }
      _this.referencedScopes.push(binding.scope);
    }
  };

  this.ThisExpression = function (path) {
    _this.seenThisExpression = true;
  };
};

function uniqueScopes(scopes) {
  return scopes.sort(function (a, b) {
    return a.uid - b.uid;
  }).reduce(function (a, x) {
    return a.length && a[a.length - 1].uid === x.uid ? a : a.concat(x);
  }, []);
}

function deepestScopeOf(path, scopes) {
  var scope = path.scope;
  do {
    if (scopes.indexOf(scope) !== -1) {
      return scope;
    }
  } while (scope = scope.parent); // eslint-disable-line no-cond-assign
}